@@ -2,8 +2,6 @@
 """
 Module with several geodetic related methods
 """
-# import math
-# import sys
 import collections
 import math
 import typing
@@ -22,6 +20,7 @@ XYZ = collections.namedtuple('XYZ', 'x y z')
 ENU = collections.namedtuple('ENU', 'east north up')
 ElAz = collections.namedtuple('ElAz', 'elevation azimuth')
 
+
 class Coordinates(object):
     """
     Class to store coordinates and transform among various coordinate systems
@@ -102,7 +101,7 @@ class Coordinates(object):
         xyz_self = self.xyz()
         xyz_other = other.xyz()
 
-        return XYZ(x=xyz_self.x-xyz_other.x, y=xyz_self.y-xyz_other.y, z=xyz_self.z-xyz_other.z)
+        return XYZ(x=xyz_self.x - xyz_other.x, y=xyz_self.y - xyz_other.y, z=xyz_self.z - xyz_other.z)
 
     @staticmethod
     def from_lla(longitude_deg: float, latitude_deg: float, height_m: float,
@@ -127,7 +126,6 @@ class Coordinates(object):
 
         return Coordinates(x, y, z)
 
-# ------------------------------------------------------------------------------
 
 class CoordinatesDisplacement(object):
     """
@@ -144,7 +142,7 @@ class CoordinatesDisplacement(object):
         self.xyz = xyz
         self.enu = enu
         self.ref_lla = None
-        
+
         if ref_position:
             self.set_reference_position(ref_position)
 
@@ -185,8 +183,8 @@ class CoordinatesDisplacement(object):
                 raise ValueError("Missing reference position in CoordinatesDisplacement to convert frame (ENU to XYZ)")
             xyz_list = enu_to_ecef(self.ref_lla.longitude, self.ref_lla.latitude, *self.enu, radians=True)
             self.xyz = XYZ(*xyz_list)
-            
-        return self.xyz 
+
+        return self.xyz
 
     def get_enu(self) -> ENU:
         """
@@ -205,7 +203,7 @@ class CoordinatesDisplacement(object):
                 raise ValueError("Missing reference position in CoordinatesDisplacement to convert frame (XYZ to ENU)")
             enu_list = ecef_to_enu(self.ref_lla.longitude, self.ref_lla.latitude, *self.xyz, radians=True)
             self.enu = ENU(*enu_list)
-            
+
         return self.enu
 
     def get_absolute_position(self) -> Coordinates:
@@ -221,13 +219,12 @@ class CoordinatesDisplacement(object):
         '0.1 0.1 0.1'
         """
 
-        xyz = lla_to_xyz(self.ref_lla.longitude, self.ref_lla.latitude, self.ref_lla.altitude, radians=True) 
+        xyz = lla_to_xyz(self.ref_lla.longitude, self.ref_lla.latitude, self.ref_lla.altitude, radians=True)
 
         dxyz = self.get_xyz()
 
         return Coordinates.from_xyz(xyz[0] + dxyz[0], xyz[1] + dxyz[1], xyz[2] + dxyz[2])
 
-
     def __abs__(self) -> float:
         """
         >>> c = CoordinatesDisplacement.from_xyz(0.1, 0.1, 0.1)
@@ -242,7 +239,6 @@ class CoordinatesDisplacement(object):
 
         return math.sqrt(sum([v * v for v in vector]))
 
-
     @staticmethod
     def from_xyz(x: float, y: float, z: float):
         """
@@ -281,8 +277,6 @@ class CoordinatesDisplacement(object):
         """
         return CoordinatesDisplacement(enu=ENU(e, n, u))
 
-# ------------------------------------------------------------------------------
-
 
 def valid_coordinates(lla=None, xyz=None):
     """
@@ -295,7 +289,7 @@ def valid_coordinates(lla=None, xyz=None):
     False
     """
 
-    ALT_THRESHOLD = 10.0e3 # meters above/under surface level
+    ALT_THRESHOLD = 10.0e3  # meters above/under surface level
 
     if lla is not None and (lla[2] > ALT_THRESHOLD or lla[2] < -ALT_THRESHOLD):
         return False
@@ -306,9 +300,7 @@ def valid_coordinates(lla=None, xyz=None):
         return False
 
     return True
-    
 
-#-------------------------------------------------------------------------------
 
 def lla_to_xyz(longitude, latitude, height, a=WGS84_A, e=WGS84_E, radians=False):
     """
@@ -324,7 +316,6 @@ def lla_to_xyz(longitude, latitude, height, a=WGS84_A, e=WGS84_E, radians=False)
     >>> [float('{0:.3f}'.format(v)) for v in xyz]
     [4163316.145, 683507.935, 4767789.479]
 
-    >>> import numpy as np
     >>> lons = np.array([9.323302567, 9.323335545, 9.323368065])
     >>> lats = np.array([48.685064919, 48.685050295, 48.685036011])
     >>> hgts = np.array([373.2428, 373.2277, 373.2078])
@@ -365,12 +356,12 @@ def lla_to_xyz(longitude, latitude, height, a=WGS84_A, e=WGS84_E, radians=False)
 
     x = nalt * coslat * coslon
     y = nalt * coslat * sinlon
-    z = ((1.0-e2) * N + height) * sinlat
+    z = ((1.0 - e2) * N + height) * sinlat
 
-    return (x, y, z)
+    return x, y, z
 
 
-def convert_decimal_units_to_dms(value:float) -> DMS:
+def convert_decimal_units_to_dms(value: float) -> DMS:
     """
     Convert from longitude/latitude to Degrees, minutes and seconds
     >>> latitude_deg = 41.528442316
@@ -383,16 +374,14 @@ def convert_decimal_units_to_dms(value:float) -> DMS:
     '2 26 3.5480400'
     """
 
-    degrees = (int)(value)
+    degrees = int(value)
     minutes_float = (value - degrees) * 60
-    minutes = (int)(minutes_float)
+    minutes = int(minutes_float)
     seconds = (minutes_float - minutes) * 60
 
     return DMS(degrees, minutes, seconds)
 
 
-#-------------------------------------------------------------------------------
-
 def xyz_to_lla(x, y, z, radians=False, a=WGS84_A, e=WGS84_E):
     """
     Convert from Cartesian XYZ-ECEF coordinates to geodetic coordinates
@@ -403,10 +392,11 @@ def xyz_to_lla(x, y, z, radians=False, a=WGS84_A, e=WGS84_E):
 
     The x, y and z values can be vectors
 
-    >>> xyz_to_lla(4807314.3520, 98057.0330, 4176767.6160)
-    (1.1685266980613551, 41.170001652314625, 173.4421242615208)
+    >>> x, y, z = xyz_to_lla(4807314.3520, 98057.0330, 4176767.6160)
+    >>> assert math.isclose(x, 1.1685266980613551)
+    >>> assert math.isclose(y, 41.170001652314625)
+    >>> assert math.isclose(z, 173.4421242615208)
 
-    >>> import numpy as np
     >>> xs = np.array([4807314.3520, 4807315.3520, 4807316.3520])
     >>> ys = np.array([98057.0330, 98058.0330, 98059.0330])
     >>> zs = np.array([4176767.6160, 4176768.6160, 4176769.6160])
@@ -431,8 +421,8 @@ def xyz_to_lla(x, y, z, radians=False, a=WGS84_A, e=WGS84_E):
     '176.2946241'
     """
 
-    a2 = a ** 2     # Squared of radius, for convenience
-    e2 = e ** 2     # Squared of eccentricity, for convenience
+    a2 = a ** 2  # Squared of radius, for convenience
+    e2 = e ** 2  # Squared of eccentricity, for convenience
 
     b = np.sqrt(a2 * (1 - e2))
     b2 = b ** 2
@@ -455,140 +445,130 @@ def xyz_to_lla(x, y, z, radians=False, a=WGS84_A, e=WGS84_E):
     alt = p / np.cos(lat) - N
 
     if radians:
-        return (lon, lat, alt)
+        return lon, lat, alt
     else:
-        return (lon * 180 / np.pi, lat * 180 / np.pi, alt)
+        return lon * 180 / np.pi, lat * 180 / np.pi, alt
 
 
+def body_to_enu_matrix(yaw, pitch, roll, radians=False):
+    """
+    Computes the matrix that transforms from Body frame to ENU
+
+    This accepts scalars or vector of scalars.
+
+    Also, the yaw, pitch and roll can be expressed as radians (radians=True)
+    or degrees (radians=False)
+    """
+
+    # Convert from degrees to radians if necessary
+    ya = yaw if radians else np.deg2rad(yaw)
+    pi = pitch if radians else np.deg2rad(pitch)
+    ro = roll if radians else np.deg2rad(roll)
+
+    # Compute the sines and cosines, used later to compute the elements
+    # of the transformation matrix
+    siny = np.sin(ya)
+    cosy = np.cos(ya)
+    sinp = np.sin(pi)
+    cosp = np.cos(pi)
+    sinr = np.sin(ro)
+    cosr = np.cos(ro)
+
+    cosy_cosr = cosy * cosr
+    siny_cosr = siny * cosr
+    sinp_sinr = sinp * sinr
+    sinr_siny = sinr * siny
+
+    # Compute the matrix coefficients
+    m = np.zeros((3, 3), dtype=object)
+
+    m[0, 0] = cosp * cosy
+    m[0, 1] = sinp_sinr * cosy - siny_cosr
+    m[0, 2] = cosy_cosr * sinp + sinr_siny
+
+    m[1, 0] = cosp * siny
+    m[1, 1] = sinp_sinr * siny + cosr * cosy
+    m[1, 2] = siny_cosr * sinp - sinr * cosy
+
+    m[2, 0] = - sinp
+    m[2, 1] = sinr * cosp
+    m[2, 2] = cosr * cosp
+
+    return m
+
+
+def body_to_enu(yaw, pitch, roll, x, y, z, radians=False, matrix=None):
+    """
+    Converts from Body reference frame to ENU reference point.
+
+    This accepts scalars or vector of scalars.
+
+    Also, the yaw, pitch and roll can be expressed as radians (radians=True)
+    or degrees (radians=False)
+
+    TEST RATIONALE HAS TO BE CHANGED ACCORDING TO THE CHANGE IN THE BODY_TO_ENU_MATRIX RESULT INTERPRETATION:
+    BODY_TO_ENU_MATRIX IS ACTUALLY IMPLEMENTING BODY2NED_MATRIX!!!!!!
+
+    >>> body_to_enu(0, 0, 0, 1, 1, 1)
+    (1.0, 1.0, -1.0)
+
+    >>> enu = body_to_enu(90, 0, 0, 1, 1, 1)
+    >>> np.round(enu)
+    array([ 1., -1., -1.])
+
+    >>> enu = body_to_enu(0, 90, 0, 1, 1, 1)
+    >>> np.round(enu)
+    array([1., 1., 1.])
+
+    >>> enu = body_to_enu(0, 0, 90, 1, 1, 1)
+    >>> np.round(enu)
+    array([-1.,  1., -1.])
+
+    >>> yaws = [0] * 3; pitches = [0] * 3; rolls = [0] * 3
+    >>> xs = [1] * 3; ys = [1] * 3; zs = [1]*3
+    >>> es, ns, us = body_to_enu(yaws, pitches, rolls, xs, ys, zs)
+    >>> es
+    array([1., 1., 1.])
+    >>> ns
+    array([1., 1., 1.])
+    >>> us
+    array([-1., -1., -1.])
+    """
+
+
+    if matrix is None:
+        m = body_to_enu_matrix(yaw, pitch, roll, radians=radians)
+
+    n = m[0, 0] * x + m[0, 1] * y + m[0, 2] * z
+    e = m[1, 0] * x + m[1, 1] * y + m[1, 2] * z
+    d = m[2, 0] * x + m[2, 1] * y + m[2, 2] * z
+
+    return e, n, -d
+
+
+def enu_to_body(yaw, pitch, roll, e, n, u, radians=False, matrix=None):
+    """
+    Converts from ENU to XYZ Body reference frame
+
+    This accepts scalars or vector of scalars.
+
+    Also, the yaw, pitch and roll can be expressed as radians (radians=True)
+    or degrees (radians=False)
+
+
+    >>> np.round(enu_to_body(0, 0, 0, *body_to_enu(0, 0, 0, 0.2, 0.2, 1.5)), decimals=1)
+    array([0.2, 0.2, 1.5])
+    """
+
+    if matrix is None:
+        m = body_to_enu_matrix(yaw, pitch, roll, radians=radians)
+
+    x = m[0, 0] * n + m[1, 0] * e + m[2, 0] * -u
+    y = m[0, 1] * n + m[1, 1] * e + m[2, 1] * -u
+    z = m[0, 2] * n + m[1, 2] * e + m[2, 2] * -u
+
+    return x, y, z
 
-# #-------------------------------------------------------------------------------
-#
-# def body2enu_matrix(yaw, pitch, roll, radians=False):
-#     """
-#     Computes the matrix that transforms from Body frame to ENU
-#
-#     This accepts scalars or vector of scalars.
-#
-#     Also, the yaw, pitch and roll can be expressed as radians (radians=True)
-#     or degrees (radians=False)
-#     """
-#
-#     # Convert from degrees to radians if necessary
-#     ya = yaw if radians else np.deg2rad(yaw)
-#     pi = pitch if radians else np.deg2rad(pitch)
-#     ro = roll if radians else np.deg2rad(roll)
-#
-#     # Compute the sines and cosines, used later to compute the elements
-#     # of the transformation matrix
-#     siny = np.sin(ya)
-#     cosy = np.cos(ya)
-#     sinp = np.sin(pi)
-#     cosp = np.cos(pi)
-#     sinr = np.sin(ro)
-#     cosr = np.cos(ro)
-#
-#     cosy_cosr = cosy * cosr
-#     siny_cosr = siny * cosr
-#     sinp_sinr = sinp * sinr
-#     sinr_siny = sinr * siny
-#
-#     # Compute the matrix coefficients
-#     m = np.zeros((3, 3), dtype=object)
-#
-#     m[0,0] = cosp * cosy
-#     m[0,1] = sinp_sinr * cosy - siny_cosr
-#     m[0,2] = cosy_cosr * sinp + sinr_siny
-#
-#     m[1,0] =  cosp * siny
-#     m[1,1] =  sinp_sinr * siny + cosr * cosy
-#     m[1,2] =  siny_cosr * sinp - sinr * cosy
-#
-#     m[2,0] = - sinp
-#     m[2,1] = sinr * cosp
-#     m[2,2] = cosr * cosp
-#
-#     return m
-#
-#
-# #-------------------------------------------------------------------------------
-#
-# def body2enu(yaw, pitch, roll, x, y, z, radians=False, matrix=None):
-#     """
-#     Converts from Body reference frame to ENU reference point.
-#
-#     This accepts scalars or vector of scalars.
-#
-#     Also, the yaw, pitch and roll can be expressed as radians (radians=True)
-#     or degrees (radians=False)
-#
-#     TEST RATIONALE HAS TO BE CHANGED ACCORDING TO THE CHANGE IN THE BODY2ENU_MATRIX RESULT INTERPRETATION:
-#     BODY2ENU_MATRIX IS ACTUALLY IMPLEMENTING BODY2NED_MATRIX!!!!!!
-#
-#     >>> body2enu(0, 0, 0, 1, 1, 1)
-#     (1.0, 1.0, -1.0)
-#
-#     >>> import numpy as np
-#
-#     >>> enu = body2enu(90, 0, 0, 1, 1, 1)
-#     >>> np.round(enu)
-#     array([ 1., -1., -1.])
-#
-#     >>> enu = body2enu(0, 90, 0, 1, 1, 1)
-#     >>> np.round(enu)
-#     array([1., 1., 1.])
-#
-#     >>> enu = body2enu(0, 0, 90, 1, 1, 1)
-#     >>> np.round(enu)
-#     array([-1.,  1., -1.])
-#
-#     >>> yaws = [0] * 3; pitches = [0] * 3; rolls = [0] * 3
-#     >>> xs = [1] * 3; ys = [1] * 3; zs = [1]*3
-#     >>> es, ns, us = body2enu(yaws, pitches, rolls, xs, ys, zs)
-#     >>> es
-#     array([1., 1., 1.])
-#     >>> ns
-#     array([1., 1., 1.])
-#     >>> us
-#     array([-1., -1., -1.])
-#     """
-#
-#     if matrix is None:
-#         m = body2enu_matrix(yaw, pitch, roll, radians=radians)
-#
-#     n = m[0,0] * x + m[0,1] * y + m[0,2] * z
-#     e = m[1,0] * x + m[1,1] * y + m[1,2] * z
-#     d = m[2,0] * x + m[2,1] * y + m[2,2] * z
-#
-#     return (e, n, -d)
-#
-# #-------------------------------------------------------------------------------
-#
-# def enu2body(yaw, pitch, roll, e, n, u, radians=False, matrix=None):
-#     """
-#     Converts from ENU to XYZ Body reference frame
-#
-#     This accepts scalars or vector of scalars.
-#
-#     Also, the yaw, pitch and roll can be expressed as radians (radians=True)
-#     or degrees (radians=False)
-#
-#
-#     >>> np.round(enu2body(0, 0, 0, *body2enu(0, 0, 0, 0.2, 0.2, 1.5)), decimals=1)
-#     array([0.2, 0.2, 1.5])
-#     """
-#
-#     if matrix is None:
-#         m = body2enu_matrix(yaw, pitch, roll, radians=radians)
-#
-#     x = m[0,0] * n + m[1,0] * e + m[2,0] * -u
-#     y = m[0,1] * n + m[1,1] * e + m[2,1] * -u
-#     z = m[0,2] * n + m[1,2] * e + m[2,2] * -u
-#
-#     return (x, y, z)
-#
-#
-#
-# -------------------------------------------------------------------------------
 
 def enu_to_ecef_matrix(longitude, latitude, radians=False):
     """
@@ -631,7 +611,6 @@ def enu_to_ecef_matrix(longitude, latitude, radians=False):
 
     return m
 
-# -------------------------------------------------------------------------------
 
 def enu_to_ecef(longitude, latitude, e, n, u, radians=False, matrix=None):
     """
@@ -679,14 +658,12 @@ def enu_to_ecef(longitude, latitude, e, n, u, radians=False, matrix=None):
     if matrix is None:
         m = enu_to_ecef_matrix(longitude, latitude, radians=radians)
 
-
     x = m[0, 0] * e + m[0, 1] * n + m[0, 2] * u
     y = m[1, 0] * e + m[1, 1] * n + m[1, 2] * u
     z = m[2, 0] * e + m[2, 1] * n + m[2, 2] * u
 
-    return (x, y, z)
+    return x, y, z
 
-# -------------------------------------------------------------------------------
 
 def xyz_to_enu(ref_pos, x, y, z, a=WGS84_A, e=WGS84_E):
     """
@@ -711,7 +688,6 @@ def xyz_to_enu(ref_pos, x, y, z, a=WGS84_A, e=WGS84_E):
 
     return ecef_to_enu(lla[0], lla[1], x, y, z, radians=True)
 
-# -------------------------------------------------------------------------------
 
 def ecef_to_enu(longitude, latitude, x, y, z, radians=False, matrix=None):
     """
@@ -745,136 +721,126 @@ def ecef_to_enu(longitude, latitude, x, y, z, radians=False, matrix=None):
     n = m[0, 1] * x + m[1, 1] * y + m[2, 1] * z
     u = m[0, 2] * x + m[1, 2] * y + m[2, 2] * z
 
-    return (e, n, u)
+    return e, n, u
 
-# #-------------------------------------------------------------------------------
-# 
-# def body2ecef(longitude, latitude, yaw, pitch, roll, x, y, z, radians=False):
-#     """
-#     Transform from body fixed reference frame to Cartesian XYZ-ECEF
-#     """
-# 
-#     enu = body2enu(yaw, pitch, roll, x, y, z, radians=radians)
-# 
-#     return enu2ecef(longitude, latitude, *enu, radians=radians)
-# 
-# 
-# #-------------------------------------------------------------------------------
-# 
-# 
-# def transposePosGeodetic(longitude, latitude, height, yaw, pitch, roll, lx, ly, lz, radians=False):
-#     """
-#     Apply lever arm (l=[lx,ly,lz]) to position (longitude,latitude, height) according to rover attitude (yaw,pitch,roll)
-#     """
-# 
-#     leverArmEnu = body2enu(yaw, pitch, roll, lx,ly, lz, radians=radians)
-#     leverArmNed = (leverArmEnu[1],leverArmEnu[0],-1*leverArmEnu[2])
-# 
-#     deltaPos =  smallPerturbationCart2Geodetic(leverArmNed[0],leverArmNed[1],leverArmNed[2],latitude,height,radians=radians)
-# 
-#     return (latitude + deltaPos[0], longitude + deltaPos[1], height + deltaPos[2])
-# 
-# 
-# #-------------------------------------------------------------------------------
-# 
-# def smallPerturbationCart2Geodetic(n, e, d ,lat, h, radians):
-#     """
-#     Transform small perturbations from cartesian (in local navigation frame NED)  to geodetic, small angle
-#     aproximation has to apply, which means norm(n,e,d) <<< Earth Radius
-#     """
-#     if not radians :
-#         lat = lat * np.pi / 180
-# 
-#     RE = WGS84_A / np.sqrt((1-(np.sin(lat)**2)*(WGS84_E**2))**3)
-#     RN = WGS84_A*(1-WGS84_E**2) / np.sqrt(1-(np.sin(lat)**2)*(WGS84_E**2))
-# 
-#     deltaLat = n / (RN+h)
-#     deltaLon = e / ((RE+h)*np.cos(lat))
-#     deltaH = -1*d
-# 
-#     if radians:
-#         return (deltaLat, deltaLon, deltaH)
-#     else:
-#         return ((deltaLat * 180 / np.pi, deltaLon * 180 / np.pi, deltaH))
-# 
-# #-------------------------------------------------------------------------------
-# 
-# def haversine(lon1, lat1, lon2, lat2, r=6371):
-#     """
-#     Calculate the great circle distance between two points
-#     on the earth (specified in decimal degrees) using the Haversine formula
-# 
-#     Return units: kilometers
-# 
-#     Extracted from http://stackoverflow.com/questions/4913349
-# 
-#     >>> haversine(0, 0, 0, 0)
-#     0.0
-# 
-#     Example extracted from https://rosettacode.org/wiki/Haversine_formula
-# 
-#     >>> haversine(-86.67, 36.12, -118.40, 33.94, r=6372.8)
-#     2887.259950607111
-#     """
-#     # convert decimal degrees to radians
-#     lon1, lat1, lon2, lat2 = map(math.radians, [lon1, lat1, lon2, lat2])
-# 
-#     # haversine formula
-#     dlon = lon2 - lon1
-#     dlat = lat2 - lat1
-#     a = math.sin(dlat/2)**2 + math.cos(lat1) * math.cos(lat2) * math.sin(dlon/2)**2
-#     c = 2 * math.asin(math.sqrt(a))
-# 
-#     return c * r
-# 
-# #-------------------------------------------------------------------------------
-# 
-# 
-# def area_triangle_in_sphere(coords, r=6371):
-#     """
-# 
-#     Area = pi*R^2*E/180
-# 
-#     R = radius of sphere
-#     E = spherical excess of triangle, E = A + B + C - 180
-#     A, B, C = angles of spherical triangle in degrees
-# 
-#     tan(E/4) = sqrt(tan(s/2)*tan((s-a)/2)*tan((s-b)/2)*tan((s-c)/2))
-# 
-#     where
-# 
-#     a, b, c = sides of spherical triangle
-#     s = (a + b + c)/2
-# 
-#     http://mathforum.org/library/drmath/view/65316.html
-# 
-#     # Compute approximate area of India (3287263 km2)
-#     >>> pWest = (68.752441, 23.483401)
-#     >>> pEast = (90.153809, 22.87744)
-#     >>> pNorth = (76.723022, 32.87036)
-#     >>> pSouth = (77.45636, 8.119053)
-#     >>> triangle1 = (pWest,pEast,pNorth)
-#     >>> triangle2 = (pWest,pEast,pSouth)
-#     >>> area_triangle_in_sphere(triangle1) + area_triangle_in_sphere(triangle2)
-#     3028215.293314756
-#     """
-# 
-#     # Compute the spherical areas using the haversine method
-#     a = haversine(coords[0][0], coords[0][1], coords[1][0], coords[1][1], r=1)
-#     b = haversine(coords[1][0], coords[1][1], coords[2][0], coords[2][1], r=1)
-#     c = haversine(coords[0][0], coords[0][1], coords[2][0], coords[2][1], r=1)
-# 
-#     s = (a + b + c) / 2.0
-# 
-#     tanE = math.sqrt(math.tan(s/2)*math.tan((s-a)/2)*math.tan((s-b)/2)*math.tan((s-c)/2))
-# 
-#     E = math.atan(tanE * 4.0)
-# 
-#     area = r*r*E
-# 
-#     return area
-# 
-# #-------------------------------------------------------------------------------
+
+def body_to_ecef(longitude, latitude, yaw, pitch, roll, x, y, z, radians=False):
+    """
+    Transform from body fixed reference frame to Cartesian XYZ-ECEF
+    """
+
+    enu = body_to_enu(yaw, pitch, roll, x, y, z, radians=radians)
+
+    return enu_to_ecef(longitude, latitude, *enu, radians=radians)
+
+
+def transpose_pos_geodetic(longitude, latitude, height, yaw, pitch, roll, lx, ly, lz, radians=False):
+    """
+    Apply lever arm (l=[lx,ly,lz]) to position (longitude,latitude, height) according to rover attitude (yaw,pitch,roll)
+    """
+
+    leverArmEnu = body_to_enu(yaw, pitch, roll, lx, ly, lz, radians=radians)
+    leverArmNed = (leverArmEnu[1], leverArmEnu[0], -1 * leverArmEnu[2])
+
+    deltaPos = small_perturbation_cart_to_geodetic(leverArmNed[0], leverArmNed[1], leverArmNed[2], latitude, height,
+                                                   radians=radians)
+
+    return latitude + deltaPos[0], longitude + deltaPos[1], height + deltaPos[2]
+
+
+def small_perturbation_cart_to_geodetic(n, e, d, lat, h, radians):
+    """
+    Transform small perturbations from cartesian (in local navigation frame NED)  to geodetic, small angle
+    aproximation has to apply, which means norm(n,e,d) <<< Earth Radius
+    """
+    if not radians:
+        lat = lat * np.pi / 180
+
+    RE = WGS84_A / np.sqrt((1 - (np.sin(lat) ** 2) * (WGS84_E ** 2)) ** 3)
+    RN = WGS84_A * (1 - WGS84_E ** 2) / np.sqrt(1 - (np.sin(lat) ** 2) * (WGS84_E ** 2))
+
+    deltaLat = n / (RN + h)
+    deltaLon = e / ((RE + h) * np.cos(lat))
+    deltaH = -1 * d
+
+    if radians:
+        return deltaLat, deltaLon, deltaH
+    else:
+        return deltaLat * 180 / np.pi, deltaLon * 180 / np.pi, deltaH
+
+
+def haversine(lon1, lat1, lon2, lat2, r=6371):
+    """
+    Calculate the great circle distance between two points
+    on the earth (specified in decimal degrees) using the Haversine formula
+
+    Return units: kilometers
+
+    Extracted from http://stackoverflow.com/questions/4913349
+
+    >>> haversine(0, 0, 0, 0)
+    0.0
+
+    Example extracted from https://rosettacode.org/wiki/Haversine_formula
+
+    >>> haversine(-86.67, 36.12, -118.40, 33.94, r=6371.8)
+    2886.8068907353736
+    """
+    # convert decimal degrees to radians
+    lon1, lat1, lon2, lat2 = map(math.radians, [lon1, lat1, lon2, lat2])
+
+    # haversine formula
+    dlon = lon2 - lon1
+    dlat = lat2 - lat1
+    a = math.sin(dlat / 2) ** 2 + math.cos(lat1) * math.cos(lat2) * math.sin(dlon / 2) ** 2
+    c = 2 * math.asin(math.sqrt(a))
+
+    return c * r
+
+
+def area_triangle_in_sphere(coords, r=6371):
+    """
+
+    Area = pi*R^2*E/180
+
+    R = radius of sphere
+    E = spherical excess of triangle, E = A + B + C - 180
+    A, B, C = angles of spherical triangle in degrees
+
+    tan(E/4) = sqrt(tan(s/2)*tan((s-a)/2)*tan((s-b)/2)*tan((s-c)/2))
+
+    where
+
+    a, b, c = sides of spherical triangle
+    s = (a + b + c)/2
+
+    http://mathforum.org/library/drmath/view/65316.html
+
+    # Compute approximate area of India (3287263 km2)
+    >>> pWest = (68.752441, 23.483401)
+    >>> pEast = (90.153809, 22.87744)
+    >>> pNorth = (76.723022, 32.87036)
+    >>> pSouth = (77.45636, 8.119053)
+    >>> triangle1 = (pWest,pEast,pNorth)
+    >>> triangle2 = (pWest,pEast,pSouth)
+    >>> area_triangle_in_sphere(triangle1) + area_triangle_in_sphere(triangle2)
+    3028215.293314756
+    """
+
+    # Compute the spherical areas using the haversine method
+    a = haversine(coords[0][0], coords[0][1], coords[1][0], coords[1][1], r=1)
+    b = haversine(coords[1][0], coords[1][1], coords[2][0], coords[2][1], r=1)
+    c = haversine(coords[0][0], coords[0][1], coords[2][0], coords[2][1], r=1)
+
+    s = (a + b + c) / 2.0
+
+    tanE = math.sqrt(math.tan(s / 2) * math.tan((s - a) / 2) * math.tan((s - b) / 2) * math.tan((s - c) / 2))
+
+    E = math.atan(tanE * 4.0)
+
+    area = r * r * E
+
+    return area
 
 
 class Dop(object):
@@ -891,14 +857,14 @@ class Dop(object):
     $GPGSV,3,2,11,02,39,223,19,13,28,070,17,26,23,252,,04,14,186,14*79
     $GPGSV,3,3,11,29,09,301,24,16,09,020,,36,,,*76
 
-    >>> elAz = [[63,137],
-    ...         [61, 98],
-    ...         [59,290],
-    ...         [39,223],
-    ...         [ 9,301],
-    ...         [14,186],
-    ...         [54,157],
-    ...         [28, 70]]
+    >>> elAz = [ElAz(63,137),
+    ...         ElAz(61, 98),
+    ...         ElAz(59,290),
+    ...         ElAz(39,223),
+    ...         ElAz( 9,301),
+    ...         ElAz(14,186),
+    ...         ElAz(54,157),
+    ...         ElAz(28, 70)]
     >>> dop = Dop(elAz)
     >>> f'{dop.pdop():.2f}'
     '1.72'
@@ -917,35 +883,34 @@ class Dop(object):
     Increasing 30 degrees the azimuth position of 3 sats (+ one in zenith), doubles the vdop value
     leaving pdop approximately constant
 
-    >>> el_az_list = [[ 0,   0],
-    ...               [ 0, 120],
-    ...               [ 0, 240],
-    ...               [90,   0]]
+    >>> el_az_list = [ElAz( 0,   0),
+    ...               ElAz( 0, 120),
+    ...               ElAz( 0, 240),
+    ...               ElAz(90,   0)]
     >>> dop = Dop(el_az_list)
     >>> f'{dop.vdop():.2f}'
     '1.15'
     >>> f'{dop.hdop():.2f}'
     '1.15'
 
-    >>> el_az_list = [[30,   0],
-    ...               [30, 120],
-    ...               [30, 240],
-    ...               [90,  0]]
+    >>> el_az_list = [ElAz(30,   0),
+    ...               ElAz(30, 120),
+    ...               ElAz(30, 240),
+    ...               ElAz(90,  0)]
     >>> dop = Dop(el_az_list)
     >>> f'{dop.vdop():.2f}'
     '2.31'
     >>> f'{dop.hdop():.2f}'
     '1.33'
     """
-    
-    def __init__(self, el_az_list:typing.List[ElAz], angles_in_degrees:bool=True) -> None:
-        
-        self.el_az_list = el_az_list   
-        self.angles_in_degrees = angles_in_degrees 
+
+    def __init__(self, el_az_list: typing.List[ElAz], angles_in_degrees: bool = True) -> None:
+
+        self.el_az_list = el_az_list
+        self.angles_in_degrees = angles_in_degrees
         self.__G()  # generate G matrix
         self.__Q()  # generate Q matrix
 
-
     def __G(self) -> None:
         """ 
         Construct geometry matrix G
@@ -958,23 +923,22 @@ class Dop(object):
             azimuths = np.radians(azimuths)
 
         n_satellites = len(self.el_az_list)
-        
+
         gx = np.cos(elevations) * np.sin(azimuths)
         gy = np.cos(elevations) * np.cos(azimuths)
         gz = np.sin(elevations)
 
-        rows = list(zip(gx, gy, gz, [1]*n_satellites))
+        rows = list(zip(gx, gy, gz, [1] * n_satellites))
 
         self.G = np.array(rows)
 
-
     def __Q(self) -> None:
         """
         Construct co-factor matrix
         """
         try:
             self.Q = np.linalg.inv(np.dot(np.matrix.transpose(self.G), self.G))
-        except np.linalg.linalg.LinAlgError as e:   
+        except np.linalg.linalg.LinAlgError as e:
             self.Q = np.full([4, 4], np.nan)
             roktools.logger.warning("Could not compute DOP. Reason {}".format(e))
 
@@ -985,7 +949,6 @@ class Dop(object):
 
         return np.sqrt(self.Q[0, 0] + self.Q[1, 1] + self.Q[2, 2] + self.Q[3, 3])
 
-
     def pdop(self) -> float:
         """
         Position DOP
@@ -993,14 +956,12 @@ class Dop(object):
 
         return np.sqrt(self.Q[0, 0] + self.Q[1, 1] + self.Q[2, 2])
 
-
     def tdop(self) -> float:
         """
         Time DOP
         """
 
         return np.sqrt(self.Q[3, 3])
-        
 
     def vdop(self) -> float:
         """
@@ -1009,7 +970,6 @@ class Dop(object):
 
         return np.sqrt(self.Q[2, 2])
 
-
     def hdop(self) -> float:
         """
         Horizontal DOP
@@ -1020,4 +980,5 @@ class Dop(object):
 
 if __name__ == "__main__":
     import doctest
+
     doctest.testmod(raise_on_error=True)
